
from amplitude_tracker.version import VERSION
from amplitude_tracker.client import Client

__version__ = VERSION

"""Settings."""
write_key = None
host = None
on_error = None
debug = False
send = True
sync_mode = False

default_client = None


def track(*args, **kwargs):
    """Send a track call."""
    _proxy('track', *args, **kwargs)


def identify(*args, **kwargs):
    """Send a identify call."""
    _proxy('identify', *args, **kwargs)


def group(*args, **kwargs):
    """Send a group call."""
    _proxy('group', *args, **kwargs)


def alias(*args, **kwargs):
    """Send a alias call."""
    _proxy('alias', *args, **kwargs)


def page(*args, **kwargs):
    """Send a page call."""
    _proxy('page', *args, **kwargs)


def screen(*args, **kwargs):
    """Send a screen call."""
    _proxy('screen', *args, **kwargs)


def flush():
    """Tell the client to flush."""
    _proxy('flush')


def join():
    """Block program until the client clears the queue"""
    _proxy('join')


def shutdown():
    """Flush all messages and cleanly shutdown the client"""
    _proxy('flush')
    _proxy('join')


def _proxy(method, *args, **kwargs):
    """Create an amplitude client if one doesn't exist and send to it."""
    global default_client
    if not default_client:
        default_client = Client(write_key, host=host, debug=debug,
                                on_error=on_error, send=send,
                                sync_mode=sync_mode)

    fn = getattr(default_client, method)
    fn(*args, **kwargs)



from contextlib import contextmanager
import functools
import unittest
import pkgutil
import logging
import sys
import json
import requests
from mock import patch, MagicMock


def all_names():
    for _, modname, _ in pkgutil.iter_modules(__path__):
        yield 'amplitude_tracker.test.' + modname


def all():
    logging.basicConfig(stream=sys.stderr)
    return unittest.defaultTestLoader.loadTestsFromNames(all_names())


@contextmanager
def patch_requests(http_code=None, response=None):
    response = response or {}
    http_code = http_code or 200
    def get_response(url, **kwargs):
        resp = requests.Response()
        resp.url = url
        resp.status_code = http_code
        resp._content = json.dumps(response).encode("utf-8")
        return resp

    with patch.multiple('requests.Session', post=MagicMock(side_effect=get_response)):
        yield {'post': requests.post}


def patch_amplitude_request(http_code=None, response=None):
    def decorator(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            with patch_requests(http_code=http_code, response=response):
                return func(*args, **kwargs)
        return inner
    return decorator



from uuid import uuid4
import time
import logging
import numbers
import atexit

from six import string_types

from amplitude_tracker.utils import clean
from amplitude_tracker.consumer import Consumer
from amplitude_tracker.request import post
from amplitude_tracker.version import VERSION

try:
    import queue
except ImportError:
    import Queue as queue


ID_TYPES = (numbers.Number, string_types)


class Client(object):
    """Create a new Amplitude client."""
    log = logging.getLogger('amplitude')

    def __init__(self, write_key=None, host=None, debug=False,
                 max_queue_size=10000, send=True, on_error=None, flush_at=100,
                 flush_interval=0.5, gzip=False, max_retries=3,
                 sync_mode=False, timeout=15, thread=1):
        require('write_key', write_key, string_types)

        self.queue = queue.Queue(max_queue_size)
        self.write_key = write_key
        self.on_error = on_error
        self.debug = debug
        self.send = send
        self.sync_mode = sync_mode
        self.host = host
        self.gzip = gzip
        self.timeout = timeout

        if debug:
            self.log.setLevel(logging.DEBUG)

        if sync_mode:
            self.consumers = None
        else:
            # On program exit, allow the consumer thread to exit cleanly.
            # This prevents exceptions and a messy shutdown when the
            # interpreter is destroyed before the daemon thread finishes
            # execution. However, it is *not* the same as flushing the queue!
            # To guarantee all messages have been delivered, you'll still need
            # to call flush().
            if send:
                atexit.register(self.join)
            for n in range(thread):
                self.consumers = []
                consumer = Consumer(
                    self.queue, write_key, host=host, on_error=on_error,
                    flush_at=flush_at, flush_interval=flush_interval,
                    gzip=gzip, retries=max_retries, timeout=timeout,
                )
                self.consumers.append(consumer)

                # if we've disabled sending, just don't start the consumer
                if send:
                    consumer.start()

    def track(self, user_id=None, event_type=None, user_properties=None,
            event_properties=None, insert_id=None, **kwargs):
        user_properties = user_properties or {}
        event_properties = event_properties or {}
        require('user_id', user_id, ID_TYPES)
        require('event_type', event_type, string_types)
        require('user_properties', user_properties, dict)
        require('event_properties', event_properties, dict)
        event_properties.update({
            'lib_version': VERSION
        })
        msg = {
            'user_id': user_id,
            'insert_id': insert_id,
            'event_type': event_type,
            'user_properties': user_properties,
            'event_properties': event_properties,
        }
        msg.update(kwargs)
        return self._enqueue(msg)

    def _enqueue(self, msg):
        """Push a new `msg` onto the queue, return `(success, msg)`"""
        msg['platform'] = 'amplitude-python'
        msg['insert_id'] = msg.get('insert_id', uuid4())
        msg['time'] = msg.get('time', int(round(time.time() * 1000)))

        require('time', msg['time'], int)
        require('event_type', msg['event_type'], string_types)
        require('user_properties', msg['user_properties'], dict)
        require('event_properties', msg['event_properties'], dict)
        msg['user_id'] = stringify_id(msg.get('user_id', None))

        msg = clean(msg)
        self.log.debug('queueing: %s', msg)

        # if send is False, return msg as if it was successfully queued
        if not self.send:
            return True, msg

        if self.sync_mode:
            self.log.debug('enqueued with blocking track.')
            post(self.write_key, timeout=self.timeout, events=[msg])

            return True, msg

        try:
            self.queue.put(msg, block=False)
            self.log.debug('message enqueued.')
            return True, msg
        except queue.Full:
            self.log.warning('amplitude queue is full')
            return False, msg

    def flush(self):
        """Forces a flush from the internal queue to the server"""
        queue = self.queue
        size = queue.qsize()
        queue.join()
        # Note that this message may not be precise, because of threading.
        self.log.debug('successfully flushed about %s items.', size)

    def join(self):
        """Ends the consumer thread once the queue is empty.
        Blocks execution until finished
        """
        for consumer in self.consumers:
            consumer.pause()
            try:
                consumer.join()
            except RuntimeError:
                # consumer thread has not started
                pass

    def shutdown(self):
        """Flush all messages and cleanly shutdown the client"""
        self.flush()
        self.join()


def require(name, field, data_type):
    """Require that the named `field` has the right `data_type`"""
    if not isinstance(field, data_type):
        msg = '{0} must have {1}, got: {2}'.format(name, data_type, field)
        raise AssertionError(msg)


def stringify_id(val):
    if val is None:
        return None
    if isinstance(val, string_types):
        return val
    return str(val)



import unittest

import amplitude_tracker


class TestModule(unittest.TestCase):

    def failed(self):
        self.failed = True

    def setUp(self):
        self.failed = False
        amplitude_tracker.write_key = 'testsecret'
        amplitude_tracker.on_error = self.failed

    def test_no_write_key(self):
        amplitude_tracker.write_key = None
        self.assertRaises(Exception, amplitude_tracker.track)

    def test_no_host(self):
        amplitude_tracker.host = None
        self.assertRaises(Exception, amplitude_tracker.track)

    def test_track(self):
        amplitude_tracker.track('userId', 'python module event')
        amplitude_tracker.flush()

    def test_flush(self):
        amplitude_tracker.flush()



import logging
from threading import Thread
import monotonic
import backoff
import json

from amplitude_tracker.request import post, APIError, DatetimeSerializer

try:
    from queue import Empty
except ImportError:
    from Queue import Empty

MAX_MSG_SIZE = 32 << 10

# Our servers only accept batches less than 500KB. Here limit is set slightly
# lower to leave space for extra data that will be added later, eg. "sentAt".
BATCH_SIZE_LIMIT = 475000


class Consumer(Thread):
    """Consumes the messages from the client's queue."""
    log = logging.getLogger('amplitude')

    def __init__(self, queue, write_key, flush_at=100, host=None,
                 on_error=None, flush_interval=0.5, gzip=False, retries=10,
                 timeout=15):
        """Create a consumer thread."""
        Thread.__init__(self)
        # Make consumer a daemon thread so that it doesn't block program exit
        self.daemon = True
        self.flush_at = flush_at
        self.flush_interval = flush_interval
        self.write_key = write_key
        self.host = host
        self.on_error = on_error
        self.queue = queue
        self.gzip = gzip
        # It's important to set running in the constructor: if we are asked to
        # pause immediately after construction, we might set running to True in
        # run() *after* we set it to False in pause... and keep running
        # forever.
        self.running = True
        self.retries = retries
        self.timeout = timeout

    def run(self):
        """Runs the consumer."""
        self.log.debug('consumer is running...')
        while self.running:
            self.upload()

        self.log.debug('consumer exited.')

    def pause(self):
        """Pause the consumer."""
        self.running = False

    def upload(self):
        """Upload the next batch of items, return whether successful."""
        success = False
        batch = self.next()
        if len(batch) == 0:
            return False

        try:
            self.request(batch)
            success = True
        except Exception as e:
            self.log.error('error uploading: %s', e)
            success = False
            if self.on_error:
                self.on_error(e, batch)
        finally:
            # mark items as acknowledged from queue
            for item in batch:
                self.queue.task_done()
            return success

    def next(self):
        """Return the next batch of items to upload."""
        queue = self.queue
        items = []

        start_time = monotonic.monotonic()
        total_size = 0

        while len(items) < self.flush_at:
            elapsed = monotonic.monotonic() - start_time
            if elapsed >= self.flush_interval:
                break
            try:
                item = queue.get(
                    block=True, timeout=self.flush_interval - elapsed)
                item_size = len(json.dumps(
                    item, cls=DatetimeSerializer).encode())
                if item_size > MAX_MSG_SIZE:
                    self.log.error(
                        'Item exceeds 32kb limit, dropping. (%s)', str(item))
                    continue
                items.append(item)
                total_size += item_size
                if total_size >= BATCH_SIZE_LIMIT:
                    self.log.debug(
                        'hit batch size limit (size: %d)', total_size)
                    break
            except Empty:
                break

        return items

    def request(self, batch):
        """Attempt to upload the batch and retry before raising an error """

        def fatal_exception(exc):
            if isinstance(exc, APIError):
                # retry on server errors and client errors
                # with 429 status code (rate limited),
                # don't retry on other client errors
                return (400 <= exc.status < 500) and exc.status != 429
            else:
                # retry on all other errors (eg. network)
                return False

        @backoff.on_exception(
            backoff.expo,
            Exception,
            max_tries=self.retries + 1,
            giveup=fatal_exception)
        def send_request():
            post(self.write_key, timeout=self.timeout, events=batch)

        send_request()



from datetime import datetime, date
import unittest
import json
import requests

from amplitude_tracker.request import post, DatetimeSerializer
from amplitude_tracker.test import patch_amplitude_request


class TestRequests(unittest.TestCase):

    @patch_amplitude_request()
    def test_valid_request(self):
        res = post('testsecret', events=[{
            'user_id': 'user_id',
            'event_type': 'python event',
        }])
        self.assertEqual(res.status_code, 200)

    def test_invalid_request_error(self):
        self.assertRaises(Exception, post, 'testsecret',
                          'https://api.amplitude.com', False, '[{]')

    def test_invalid_host(self):
        self.assertRaises(Exception, post, 'testsecret',
                          'api.amplitude.com/', events=[])

    def test_datetime_serialization(self):
        data = {'created': datetime(2012, 3, 4, 5, 6, 7, 891011)}
        result = json.dumps(data, cls=DatetimeSerializer)
        self.assertEqual(result, '{"created": "2012-03-04T05:06:07.891011"}')

    def test_date_serialization(self):
        today = date.today()
        data = {'created': today}
        result = json.dumps(data, cls=DatetimeSerializer)
        expected = '{"created": "%s"}' % today.isoformat()
        self.assertEqual(result, expected)

    @patch_amplitude_request()
    def test_should_not_timeout(self):
        res = post('testsecret', events=[{
            'user_id': 'user_id',
            'event_type': 'python event',
        }], timeout=15)
        self.assertEqual(res.status_code, 200)

    def test_should_timeout(self):
        with self.assertRaises((requests.ConnectTimeout, requests.ReadTimeout)):
            post('testsecret', events=[{
                'user_id': 'user_id',
                'event_type': 'python event',
            }], timeout=0.0001)



from datetime import date, datetime, timedelta
from decimal import Decimal
import unittest

from dateutil.tz import tzutc
import six

from amplitude_tracker import utils


class TestUtils(unittest.TestCase):

    def test_timezone_utils(self):
        now = datetime.now()
        utcnow = datetime.now(tz=tzutc())
        self.assertTrue(utils.is_naive(now))
        self.assertFalse(utils.is_naive(utcnow))

        fixed = utils.guess_timezone(now)
        self.assertFalse(utils.is_naive(fixed))

        shouldnt_be_edited = utils.guess_timezone(utcnow)
        self.assertEqual(utcnow, shouldnt_be_edited)

    def test_clean(self):
        simple = {
            'decimal': Decimal('0.142857'),
            'unicode': six.u('woo'),
            'date': datetime.now(),
            'long': 200000000,
            'integer': 1,
            'float': 2.0,
            'bool': True,
            'str': 'woo',
            'none': None
        }

        complicated = {
            'exception': Exception('This should show up'),
            'timedelta': timedelta(microseconds=20),
            'list': [1, 2, 3]
        }

        combined = dict(simple.items())
        combined.update(complicated.items())

        pre_clean_keys = combined.keys()

        utils.clean(combined)
        self.assertEqual(combined.keys(), pre_clean_keys)

    def test_clean_with_dates(self):
        dict_with_dates = {
            'birthdate': date(1980, 1, 1),
            'registration': datetime.utcnow(),
        }
        self.assertEqual(dict_with_dates, utils.clean(dict_with_dates))

    def test_bytes(self):
        if six.PY3:
            item = bytes(10)
        else:
            item = bytearray(10)

        utils.clean(item)

    def test_clean_fn(self):
        cleaned = utils.clean({'fn': lambda x: x, 'number': 4})
        self.assertEqual(cleaned['number'], 4)
        # TODO: fixme, different behavior on python 2 and 3
        if 'fn' in cleaned:
            self.assertEqual(cleaned['fn'], None)

    def test_remove_slash(self):
        self.assertEqual('http://api.amplitude.com',
                         utils.remove_trailing_slash('http://api.amplitude.com/'))
        self.assertEqual('http://api.amplitude.com',
                         utils.remove_trailing_slash('http://api.amplitude.com'))



from datetime import date
import unittest
import six
import mock
import time

from amplitude_tracker.version import VERSION
from amplitude_tracker.client import Client
from amplitude_tracker.test import patch_amplitude_request


class TestClient(unittest.TestCase):

    def fail(self, e, batch):
        """Mark the failure handler"""
        self.failed = True

    def setUp(self):
        self.failed = False
        self.client = Client('testsecret', on_error=self.fail)

    def test_requires_write_key(self):
        self.assertRaises(AssertionError, Client)

    def test_empty_flush(self):
        self.client.flush()

    @patch_amplitude_request()
    def test_basic_track(self):
        client = self.client
        success, msg = client.track(
            user_id='userId', event_type='python test event')
        client.flush()
        self.assertTrue(success)
        self.assertFalse(self.failed)

        self.assertEqual(msg['event_type'], 'python test event')
        self.assertTrue(isinstance(msg['time'], int))
        self.assertEqual(msg['user_id'], 'userId')
        self.assertEqual(msg['user_properties'], {})
        self.assertEqual(msg['event_properties'], {'lib_version': VERSION})

    @patch_amplitude_request()
    def test_stringifies_user_id(self):
        # A large number that loses precision in node:
        # node -e "console.log(157963456373623802 + 1)" > 157963456373623800
        client = self.client
        success, msg = client.track(
            user_id=157963456373623802, event_type='python test event')
        client.flush()
        self.assertTrue(success)
        self.assertFalse(self.failed)

        self.assertEqual(msg['user_id'], '157963456373623802')

    @patch_amplitude_request()
    def test_advanced_track(self):
        client = self.client
        success, msg = client.track(
            user_id='userId',
            event_type='python test event',
            user_properties={'property': 'value'},
            ip='192.168.0.1',
            time=1234,
            insert_id='insertId')

        self.assertTrue(success)

        self.assertEqual(msg['time'], 1234)
        self.assertEqual(msg['user_properties'], {'property': 'value'})
        self.assertEqual(msg['event_properties'], {'lib_version': VERSION})
        self.assertEqual(msg['ip'], '192.168.0.1')
        self.assertEqual(msg['event_type'], 'python test event')
        self.assertEqual(msg['insert_id'], 'insertId')
        self.assertEqual(msg['user_id'], 'userId')
        self.assertEqual(msg['platform'], 'amplitude-python')

    @patch_amplitude_request()
    def test_flush(self):
        client = self.client
        # set up the consumer with more requests than a single batch will allow
        for i in range(1000):
            success, msg = client.track(
                user_id=157963456373623803, event_type='python test event')
        # We can't reliably assert that the queue is non-empty here; that's
        # a race condition. We do our best to load it up though.
        client.flush()
        # Make sure that the client queue is empty after flushing
        self.assertTrue(client.queue.empty())

    @patch_amplitude_request()
    def test_shutdown(self):
        client = self.client
        # set up the consumer with more requests than a single batch will allow
        for i in range(1000):
            success, msg = client.track(
                user_id=157963456373623803, event_type='python test event')
        client.shutdown()
        # we expect two things after shutdown:
        # 1. client queue is empty
        # 2. consumer thread has stopped
        self.assertTrue(client.queue.empty())
        for consumer in client.consumers:
            self.assertFalse(consumer.is_alive())

    @patch_amplitude_request()
    def test_synchronous(self):
        client = Client('testsecret', sync_mode=True)

        success, message = client.track(
            user_id=157963456373623803, event_type='python test event')
        self.assertFalse(client.consumers)
        self.assertTrue(client.queue.empty())
        self.assertTrue(success)

    @patch_amplitude_request()
    def test_overflow(self):
        client = Client('testsecret', max_queue_size=1)
        # Ensure consumer thread is no longer uploading
        client.join()

        for i in range(10):
            client.track(
                user_id=157963456373623803, event_type='python test event')

        success, msg = client.track(
                user_id=157963456373623803, event_type='python test event')
        # Make sure we are informed that the queue is at capacity
        self.assertFalse(success)

    @patch_amplitude_request()
    def test_success_on_invalid_write_key(self):
        client = Client('bad_key', on_error=self.fail)
        client.track('user_id', 'event_type')
        client.flush()
        self.assertFalse(self.failed)

    def test_unicode(self):
        Client(six.u('unicode_key'))

    @patch_amplitude_request()
    def test_numeric_user_id(self):
        self.client.track(1234, 'python event')
        self.client.flush()
        self.assertFalse(self.failed)

    def test_debug(self):
        Client('bad_key', debug=True)

    @patch_amplitude_request()
    def test_identify_with_date_object(self):
        client = self.client
        success, msg = client.track(
            user_id=157963456373623803,
            event_type='python test event',
            user_properties={'birthdate': date(1981, 2, 2)})
        client.flush()
        self.assertTrue(success)
        self.assertFalse(self.failed)

        self.assertEqual(
            msg['user_properties'], {'birthdate': date(1981, 2, 2)})

    @patch_amplitude_request()
    def test_gzip(self):
        client = Client('testsecret', on_error=self.fail, gzip=True)
        for _ in range(10):
            client.track(
                user_id=157963456373623803,
                event_type='python test event',
                user_properties={'trait': 'value'})
        client.flush()
        self.assertFalse(self.failed)

    @patch_amplitude_request()
    def test_user_defined_flush_at(self):
        client = Client('testsecret', on_error=self.fail,
                        flush_at=10, flush_interval=3)

        def mock_post_fn(*args, **kwargs):
            self.assertEqual(len(kwargs['batch']), 10)

        # the post function should be called 2 times, with a batch size of 10
        # each time.
        with mock.patch('amplitude_tracker.consumer.post', side_effect=mock_post_fn) \
                as mock_post:
            for _ in range(20):
                client.track(
                    user_id=157963456373623803,
                    event_type='python test event',
                    user_properties={'trait': 'value'})
            time.sleep(1)
            self.assertEqual(mock_post.call_count, 2)

    @patch_amplitude_request()
    def test_user_defined_timeout(self):
        client = Client('testsecret', timeout=10)
        for consumer in client.consumers:
            self.assertEqual(consumer.timeout, 10)

    @patch_amplitude_request()
    def test_default_timeout_15(self):
        client = Client('testsecret')
        for consumer in client.consumers:
            self.assertEqual(consumer.timeout, 15)



from datetime import date, datetime
import logging
import json
from requests import sessions

from amplitude_tracker.version import VERSION

_session = sessions.Session()


def post(write_key, timeout=15, **kwargs):
    """Post the `kwargs` to the API"""
    log = logging.getLogger('amplitude')
    body = kwargs
    url = 'https://api.amplitude.com/2/httpapi'
    body.update({
        'api_key': write_key,
        'options': {'min_id_length': 0}
    })
    log.debug('making request: %s', body)
    data = json.dumps(body, cls=DatetimeSerializer)
    res = _session.post(url, data=data, timeout=timeout)

    if res.status_code == 200:
        log.debug('data uploaded successfully')
        return res

    try:
        payload = res.json()
        log.debug('received response: %s', payload)
        raise APIError(res.status_code, payload['code'], payload['error'])
    except ValueError:
        raise APIError(res.status_code, 'unknown', res.text)


class APIError(Exception):

    def __init__(self, status, code, message):
        self.message = message
        self.status = status
        self.code = code

    def __str__(self):
        msg = "[Amplitude] {0}: {1} ({2})"
        return msg.format(self.code, self.message, self.status)


class DatetimeSerializer(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (date, datetime)):
            return obj.isoformat()

        return json.JSONEncoder.default(self, obj)




import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

# Don't import analytics-python module here, since deps may not be installed
sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'amplitude_tracker'))
from version import VERSION

long_description = '''
Amplitude Tracker
=================

Amplitude Tracker library lets you record analytics data from your
Python code to `Amplitude`_

Getting Started
---------------

Install ``amplitude-tracker`` using pip:

::

   pip install amplitude-tracker

Inside your app, you’ll want to *set your* ``write_key`` before making
any analytics calls:

.. code:: python

   import amplitude_tracker as amplitude

   amplitude.write_key = 'xxxxxxxxxxxxxxx'

*Note:* If you need to send data to multiple Segment sources, you can
initialize a new Client for each write_key.

Development Settings
--------------------

The default initialization settings are production-ready and queue
messages to be processed by a background thread.

In development you might want to enable some settings to make it easier
to spot problems. Enabling amplitude.debug will log debugging info to
the Python logger. You can also add an on_error handler to specifically
print out the response you’re seeing from the Amplitude’s API.

.. code:: python

   def on_error(error, items):
       print("An error occurred:", error)


   analytics.debug = True
   analytics.on_error = on_error

Track
-----

``track`` lets you record the actions your users perform. Every action
triggers what we call an “event”, which can also have associated
properties.

.. code:: python

   import amplitude_tracker as amplitude
   amplitude.write_key = 'xxxxxxxxxxxxxxx'

   amplitude.track(
       user_id="xxx",
       event_type="xxx",
       user_properties={"trait": "xxx"},
       event_properties={"attribute": "xxx"})

Batching
--------

This library is built to support high performance environments. That
means it is safe to use amplitude-tracker on a web server that’s serving
hundreds of requests per second.

Every call ``track`` method *does not* result in an HTTP request, but is
queued in memory instead. Messages are flushed in batch in the
background, which allows for much faster operation.

By default, this library will flush:

-  every ``100`` messages (control with ``upload_size``)
-  if ``0.5`` seconds has passed since the last flush (control with
   ``upload_interval``)

There is a maximum of ``500KB`` per batch request and ``32KB`` per call.

.. _Amplitude: https://amplitude.com
'''

install_requires = [
    "requests>=2.7,<3.0",
    "six>=1.5",
    "monotonic>=1.5",
    "backoff==1.10.0",
    "python-dateutil>2.1"
]

tests_require = [
    "mock==2.0.0",
    "pylint==1.9.3",
    "flake8==3.7.9",
    "coverage==4.5.4"
]

setup(
    name='amplitude-tracker',
    version=VERSION,
    url='https://github.com/RandomCoffee/amplitude',
    author='RandomCoffee',
    author_email='tech@random-coffee.com',
    maintainer='RandomCoffee',
    maintainer_email='tech@random-coffee.com',
    test_suite='amplitude_tracker.test.all',
    packages=['amplitude_tracker', 'amplitude_tracker.test'],
    license='MIT License',
    install_requires=install_requires,
    extras_require={
        'test': tests_require
    },
    description='The hassle-free way to integrate amplitude into any python application.',
    long_description=long_description,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



import unittest
import mock
import time
import json

try:
    from queue import Queue
except ImportError:
    from Queue import Queue

from amplitude_tracker.consumer import Consumer, MAX_MSG_SIZE
from amplitude_tracker.request import APIError
from amplitude_tracker.test import patch_amplitude_request


class TestConsumer(unittest.TestCase):

    def test_next(self):
        q = Queue()
        consumer = Consumer(q, '')
        q.put(1)
        next = consumer.next()
        self.assertEqual(next, [1])

    def test_next_limit(self):
        q = Queue()
        flush_at = 50
        consumer = Consumer(q, '', flush_at)
        for i in range(10000):
            q.put(i)
        next = consumer.next()
        self.assertEqual(next, list(range(flush_at)))

    def test_dropping_oversize_msg(self):
        q = Queue()
        consumer = Consumer(q, '')
        oversize_msg = {'m': 'x' * MAX_MSG_SIZE}
        q.put(oversize_msg)
        next = consumer.next()
        self.assertEqual(next, [])
        self.assertTrue(q.empty())

    @patch_amplitude_request()
    def test_upload(self):
        q = Queue()
        consumer = Consumer(q, 'testsecret')
        track = {
            'event_type': 'python event',
            'user_id': 'userId',
        }
        q.put(track)
        success = consumer.upload()
        self.assertTrue(success)

    def test_flush_interval(self):
        # Put _n_ items in the queue, pausing a little bit more than
        # _flush_interval_ after each one.
        # The consumer should upload _n_ times.
        q = Queue()
        flush_interval = 0.3
        consumer = Consumer(q, 'testsecret', flush_at=10,
                            flush_interval=flush_interval)
        with mock.patch('amplitude_tracker.consumer.post') as mock_post:
            consumer.start()
            for i in range(0, 3):
                track = {
                    'event_type': 'python event %d' % i,
                    'user_id': 'userId'
                }
                q.put(track)
                time.sleep(flush_interval * 1.1)
            self.assertEqual(mock_post.call_count, 3)

    def test_multiple_uploads_per_interval(self):
        # Put _flush_at*2_ items in the queue at once, then pause for
        # _flush_interval_. The consumer should upload 2 times.
        q = Queue()
        flush_interval = 0.5
        flush_at = 10
        consumer = Consumer(q, 'testsecret', flush_at=flush_at,
                            flush_interval=flush_interval)
        with mock.patch('amplitude_tracker.consumer.post') as mock_post:
            consumer.start()
            for i in range(0, flush_at * 2):
                track = {
                    'event_type': 'python event %d' % i,
                    'user_id': 'userId'
                }
                q.put(track)
            time.sleep(flush_interval * 1.1)
            self.assertEqual(mock_post.call_count, 2)

    @patch_amplitude_request()
    def test_request(self):
        consumer = Consumer(None, 'testsecret')
        track = {
            'event_type': 'python event',
            'user_id': 'userId'
        }
        consumer.request([track])

    def _test_request_retry(self, consumer,
                            expected_exception, exception_count):

        def mock_post(*args, **kwargs):
            mock_post.call_count += 1
            if mock_post.call_count <= exception_count:
                raise expected_exception
        mock_post.call_count = 0

        with mock.patch('amplitude_tracker.consumer.post',
                        mock.Mock(side_effect=mock_post)):
            track = {
                'event_type': 'python event',
                'user_id': 'userId'
            }
            # request() should succeed if the number of exceptions raised is
            # less than the retries paramater.
            if exception_count <= consumer.retries:
                consumer.request([track])
            else:
                # if exceptions are raised more times than the retries
                # parameter, we expect the exception to be returned to
                # the caller.
                try:
                    consumer.request([track])
                except type(expected_exception) as exc:
                    self.assertEqual(exc, expected_exception)
                else:
                    self.fail(
                        "request() should raise an exception if still failing "
                        "after %d retries" % consumer.retries)

    def test_request_retry(self):
        # we should retry on general errors
        consumer = Consumer(None, 'testsecret')
        self._test_request_retry(consumer, Exception('generic exception'), 2)

        # we should retry on server errors
        consumer = Consumer(None, 'testsecret')
        self._test_request_retry(consumer, APIError(
            500, 'code', 'Internal Server Error'), 2)

        # we should retry on HTTP 429 errors
        consumer = Consumer(None, 'testsecret')
        self._test_request_retry(consumer, APIError(
            429, 'code', 'Too Many Requests'), 2)

        # we should NOT retry on other client errors
        consumer = Consumer(None, 'testsecret')
        api_error = APIError(400, 'code', 'Client Errors')
        try:
            self._test_request_retry(consumer, api_error, 1)
        except APIError:
            pass
        else:
            self.fail('request() should not retry on client errors')

        # test for number of exceptions raise > retries value
        consumer = Consumer(None, 'testsecret', retries=3)
        self._test_request_retry(consumer, APIError(
            500, 'code', 'Internal Server Error'), 3)

    def test_pause(self):
        consumer = Consumer(None, 'testsecret')
        consumer.pause()
        self.assertFalse(consumer.running)

    def test_max_batch_size(self):
        q = Queue()
        consumer = Consumer(
            q, 'testsecret', flush_at=100000, flush_interval=3)
        track = {
            'event_type': 'python event',
            'user_id': 'userId'
        }
        msg_size = len(json.dumps(track).encode())
        # number of messages in a maximum-size batch
        n_msgs = int(475000 / msg_size)

        def mock_post_fn(_, data, **kwargs):
            res = mock.Mock()
            res.status_code = 200
            self.assertTrue(len(data.encode()) < 500000,
                            'batch size (%d) exceeds 500KB limit'
                            % len(data.encode()))
            return res

        with mock.patch('amplitude_tracker.request._session.post',
                        side_effect=mock_post_fn) as mock_post:
            consumer.start()
            for _ in range(0, n_msgs + 2):
                q.put(track)
            q.join()
            self.assertEqual(mock_post.call_count, 2)
